from sklearn.linear_model import LogisticRegression
from sklearn.cross_validation import train_test_split
from sklearn.metrics import (precision_score, recall_score)
from mongokit import Connection
from models import (Paper, SynthTrainParagraph)
from bson.objectid import ObjectId
from nltk import (word_tokenize, pos_tag)
from nltk.stem.wordnet import WordNetLemmatizer
from csv import reader
from re import sub
import xml.etree.ElementTree as et

class SynthParaClassifier(object):

  def __init__(self):
    self.lr = LogisticRegression()
    self.fc = FeatureComputer()
    self.connection = Connection()
    self.connection.register([Paper, SynthTrainParagraph])

  def load_training_data(self):
    self.training_methods_X = []
    self.training_methods_Y = []
    pos_count_total = 0 #number of synth paragraphs (positives) in training set
    print 'Loading training data...'
    print

    all_training_methods = list(self.connection.predsynth.synth_train_paragraphs.find())
    for doc in all_training_methods:
      feature_vector = self.fc.vectorize_sections([doc['paragraph']])[0]
      if len(feature_vector) > 0 and doc['is_para_method'] is not None:
        self.training_methods_X.append(feature_vector)
        self.training_methods_Y.append(int(doc['is_para_method']))
        if doc['is_para_method']:
          pos_count_total += 1

    print 'Loaded training samples: ' + str(len(self.training_methods_Y)) + ' total and ' + str(pos_count_total) + ' positive'
    print 'Number of paragraphs in training set: ' + str(len(all_training_methods))

    self.train_pos_frac = round(float(pos_count_total) / len(self.training_methods_Y) , 4)
    print 'True positive fraction in training set: ' + str(self.train_pos_frac)

  def train(self, split):
    self.testing_methods_X = []
    self.testing_methods_Y = []

    print 'Splitting data into training/testing: ' + str(1-split) + '/' + str(split) + '... '
    self.training_methods_X, self.testing_methods_X, self.training_methods_Y, self.testing_methods_Y = train_test_split(self.training_methods_X, self.training_methods_Y, test_size=split)

    print 'Training logistic regression classifier... '
    print
    self.lr.fit(self.training_methods_X, self.training_methods_Y)

  def test(self):
    print 'Testing LR classifier accuracy...'

    self.accuracy = round(self.lr.score(self.testing_methods_X, self.testing_methods_Y), 4)
    self.adj_accuracy = round(self.accuracy / (1.0 - self.train_pos_frac), 4)

    pred_testing_methods_Y = self.lr.predict(self.testing_methods_X)
    self.precision = round(precision_score(self.testing_methods_Y, pred_testing_methods_Y), 2)
    self.recall = round(recall_score(self.testing_methods_Y, pred_testing_methods_Y), 2)

    print 'Accuracy is ' + str(self.accuracy)
    print 'Adjusted accuracy (acc/zero_guess_strat_acc) is ' + str(self.adj_accuracy)
    print 'Precision is ' + str(self.precision)
    print 'Recall is ' + str(self.recall)
    print

  def predict(self, collection='papers'):
    fails = 0
    num_papers = 0

    print 'Loading data to predict...'
    papers_to_predict = self.connection.predsynth[collection].find()
    for raw_doc in papers_to_predict:
      doc = self.connection.predsynth[collection].Paper(raw_doc)
      if doc['para_features']:
        doc['is_para_methods'] = []
        for i, paragraph_features in enumerate(doc['para_features']):
          doc['is_para_methods'].extend(list(self.lr.predict(paragraph_features)))
        doc.save()
      else:
        fails += 1
      num_papers += 1

    print 'Failed to predict/save ' + str(fails) + ' papers (no paragraphs detected in these papers)'
    print 'Predicted and saved ' + str(num_papers - fails) + ' total papers'



class SynthParaTrainer(object):

  def __init__(self):
    self.connection = Connection()
    self.connection.register([Paper])

  def load_paper_ids(self, collection):
    self.paper_ids = []
    for file in collection.find():
      self.paper_ids.append(file['_id'])

  def load_db_paper(self, collection, paper_id):
    return collection.find_one({"_id": ObjectId(paper_id)})

  def train_from_file(self, file, delim=','):
    with open(file, 'rU') as f:
      csvreader = reader(f, delimiter=delim)
      curr_paper_id = None
      curr_paper_is_methods = []

      for row in csvreader:
        paper_id = row[0]

        if row[2] == '1':
          is_method = 1
        else:
          is_method = 0

        if not paper_id == curr_paper_id:
          if curr_paper_id is not None:
            raw_doc = self.load_db_paper(self.connection.predsynth.papers, curr_paper_id)
            doc = self.connection.predsynth.t_papers.Paper(raw_doc)
            if len(doc['paragraphs']) == len(curr_paper_is_methods):
              doc['is_para_methods'] = curr_paper_is_methods
              doc.save()
              print 'Tagged and saved! Paper: ' + paper_id
            else:
              print 'Section length mismatch! Paper: ' + paper_id
              print 'Non-tagged: ' + str(len(doc['paragraphs'])) + ' and Tagged: ' + str(len(curr_paper_is_methods))

          curr_paper_is_methods = []
          curr_paper_id = paper_id


        curr_paper_is_methods.append(is_method)

  def train_loop(self):
    self.load_paper_ids(self.connection.predsynth.papers)
    did_skip_to_id = False

    for paper_id in self.paper_ids:
      skipped = False

      if did_skip_to_id:
        if not str(paper_id) == str(skip_id):
          skipped = True
        else:
          did_skip_to_id = False

      print 'Training paper ID: ' + str(paper_id)
      print

      raw_doc = self.load_db_paper(self.connection.predsynth.papers, paper_id)
      paper_doc = self.connection.predsynth.t_papers.Paper(raw_doc)

      try:
        is_methods = [0]*len(paper_doc['paragraphs'])
      except KeyError:
        print 'Failed to locate paper sections. Skipping to next paper...'
        print
        skipped = True

      if not skipped:
        for i, section in enumerate(paper_doc['paragraphs']):
          print section
          print

          try:
            resp = int(raw_input('Methods? (Y=1 N=0 Q=-1 Skip=2 SkipToID=3) : '))
            print

          except:
            print 'Could not parse response! Moving to next section...'
            print


          if resp == 0 or resp == 1:
            is_methods[i] = resp
          elif resp == -1:
            quit()
          elif resp == 2:
            skipped = True
            print 'Skipping to next paper...'
            print
            break
          elif resp == 3:
            did_skip_to_id = True
            skip_id = str(raw_input('Which ID? '))
            print
            break
          else:
            print 'Invalid response! Moving to next section...'
            print

      if not skipped:
        paper_doc['is_para_methods'] = is_methods
        paper_doc.save()


class FeatureComputer(object):

  def __init__(self):
    self.paper_ids = []
    self.connection = Connection()
    self.connection.register([Paper])

  def load_keywords(self, keywords):
    self.keywords = list(keywords)

  def load_paper_ids(self, collection):
    for file in collection.find():
      self.paper_ids.append(file['_id'])

  def load_db_paper(self, collection, paper_id):
    return collection.find_one({"_id": ObjectId(paper_id)})

  def tag_and_save_xml(self, paper_id, collection='papers'):
    paper_sections = []
    title = None
    abstract = None

    try:
      doc = self.connection.predsynth[collection].Paper(self.load_db_paper(self.connection.predsynth[collection], paper_id))
      xml = et.fromstring(doc['tagged_xml'])
      for child in xml[0][0]:
        if child.tag == 'title':
          title = unicode(child.text)
        elif child.tag == 'abstract':
          abstract = unicode(child.text)
      for child in xml[0][1]:
        if child.tag == 'paragraph-marker':
          paper_sections.append(unicode(child.text))

      doc['paragraphs'] = paper_sections
      doc['para_features'] = self.vectorize_sections(paper_sections)
      doc['title'] = title
      doc['abstract'] = abstract

      try:
        doc.save()
      except:
        pass #Document too big to save

    except et.ParseError:
      pass
    except UnicodeEncodeError:
      pass

  def vectorize_sections(self, paper_sections):
    section_vecs = []

    for section in paper_sections:
      try:
        curr_section_vec = [0]*len(self.keywords)

        for word in word_tokenize(section):
          #Clean up the word
          word = word.lower()
          word = sub('[\(\)\[\].,?\-\/:<>#=]', '', word)

          #Lemmatize (assume both noun and verb; this may be buggy!)
          lm = WordNetLemmatizer()
          word = lm.lemmatize(word, 'n')
          word = lm.lemmatize(word, 'v')

          if word in self.keywords:
            curr_section_vec[self.keywords.index(word)] += 1

        section_vecs.append(curr_section_vec)
      except TypeError:
        #Something was wrong w/ parsing the section
        pass

    return section_vecs


class KeywordScraper(object):

  def __init__(self):
    self.connection = Connection()
    self.connection.register([SynthTrainParagraph])
    self.keywords = {}
    self.training_methods = []
    self.highest_keywords = {}

  def load_training_paragraphs(self):
    for doc in self.connection.predsynth.synth_train_paragraphs.find():
      self.training_methods.append(doc['paragraph'])

  def find_keywords(self, count=1):
    for method in self.training_methods:
      try:
        for word in pos_tag(word_tokenize(method)):


          #Clean up the word
          clean_word = word[0].lower()
          clean_word = sub('[\(\)\[\].,?\-\/:<>#=]', '', clean_word)

          #Lemmatize (assume both noun and verb; this may be buggy!)
          lm = WordNetLemmatizer()
          clean_word = lm.lemmatize(clean_word, 'n')
          clean_word = lm.lemmatize(clean_word, 'v')

          if len(clean_word) > 3 and (word[1] == 'NN' or word[1] == 'NNP' or word[1] == 'VB' or word[1] == 'VBP'):
            if clean_word in self.keywords:
              self.keywords[clean_word] += count
            else:
              self.keywords[clean_word] = 1

      except UnicodeDecodeError:
        print 'Failed to decode line!'

  def find_highest_keywords(self, num_highest):
    highest_keywords = []

    for i, word in enumerate(sorted(self.keywords, key=self.keywords.get, reverse=True)):
      if i < num_highest:
        highest_keywords.append(str(word))
      else:
        return highest_keywords
