from re import match
from json import loads
from nltk.stem.wordnet import WordNetLemmatizer
from nltk import (sent_tokenize, word_tokenize, pos_tag)
from mongokit import Connection
from models import Paper
from pyfactorie.utilities import (Node, NodeHandler)
from pyfactorie.parsers import FactorieParser

class SynthRouteExtractor(object):

  def __init__(self):
    self.lm = WordNetLemmatizer()
    self.connection = Connection()
    self.connection.register([Paper])
    self.parser = FactorieParser()

    with open('data/extraction_lexicon.json') as f:
      self.lexicon = loads(f.read())

    #Known operations to match
    self.known_operation_categories = self.lexicon['known_operation_categories']
    self.known_operations = [] #Build flat list from category list
    self.known_operations_categories = [] #Build tuple list from category list
    for category in self.known_operation_categories:
      for operation in self.known_operation_categories[category]:
        self.known_operations.append(self.lm.lemmatize(operation, 'v'))
        self.known_operations_categories.append((category, self.lm.lemmatize(operation, 'v')))

    #Operating conditions
    self.known_conditions = []
    self.known_environment_conditions = self.lexicon['known_environment_conditions']
    self.known_numeric_conditions = self.lexicon['known_numeric_conditions']
    self.known_conditions.extend(self.known_environment_conditions)

    #Entity exceptions (common nouns that are not entities)
    self.known_entity_exceptions = self.lexicon['known_entity_exceptions']

    #Signal verbs for target materials
    self.known_target_material_verbs = self.lexicon['known_target_material_verbs']
    self.known_target_material_nouns = self.lexicon['known_target_material_nouns']

    print self.__class__.__name__ + ' initialized'

  def load_database_chemicals(self):
    self.database_chemicals = []
    loaded_chemicals = self.connection.entities.chemicals.find()
    for chemical in loaded_chemicals:
      self.database_chemicals.append(str(chemical['pretty_formula']))

    print "Loaded " + str(len(self.database_chemicals)) + " chemical formulae"

  def load_all_papers(self, limit, collection='papers'):
    self.papers = []
    papers = self.connection.predsynth[collection].find().limit(limit)
    for paper in papers:
      self.papers.append(self.connection.predsynth[collection].Paper(paper))
    print 'Loaded all papers!'

  def save_all_papers(self):
    for doc in self.papers:
      doc.save()
    print 'Saved all papers!'

  def extract_sentence_data(self):
    for doc in self.papers:
      doc['extracted_data'] = []
      for i, tree_list in enumerate(doc['parse_trees']):
        doc['extracted_data'].append([])
        for tree in tree_list:
          doc['extracted_data'][i].append([])
          self.extract_from_tree(tree, tree, doc['extracted_data'][i][-1])
          doc.sort_all_operations()

      del doc['parse_trees'] #shouldn't need the trees after this point

    print 'Extracted data from all sentences'

  def extract_target_materials(self):
    for doc in self.papers:
      doc['target_materials'] = self.extract_target_material_from_abstract(str(doc['abstract']))
      # extracted_targets = []
      # roots = self.parser.parse_sentence(doc['abstract'])
      # for root in roots:
      #   self.extract_target_material_from_parsed_abstract(root, extracted_targets)
      # doc['target_materials'] = extracted_targets
    print 'Extracted all target materials'

  def parse_all_sentences(self):
    for doc in self.papers:
      doc['parse_trees'] = []
      for i, paragraph in enumerate(doc['paragraphs']):
        if doc['is_para_methods'][i]:
          roots = self.parser.parse_sentence(str(paragraph))
          if not roots == 1: #if parse didn't fail
            doc['parse_trees'].append(roots)

  def extract_from_tree(self, true_root, tree, extracted_data_list):
    root = tree
    extracted_data = {
      'id': 0,
      'sent_id': 0,
      'operation': '',
      'operation_category': '',
      'entities': [],
      'conditions': []
    }

    lem_root_v = self.lm.lemmatize(str(root).lower(), 'v')
    handler = NodeHandler()

    allowed_v_pos = ['VB', 'VBN', 'VBD', 'VBG']

    #Start by getting operations and simple metadata (e.g. ID)
    if (lem_root_v in self.known_operations and
    root.data['pos'] in allowed_v_pos):
      extracted_data['operation'] = unicode(lem_root_v)
      extracted_data['id'] = int(root.data['id'])
      extracted_data['sent_id'] = int(root.data['sent_id'])
      extracted_data['operation_category'] = unicode([tup[0] for tup in self.known_operations_categories if lem_root_v == tup[1]][0])

      #Search for entities asociated with this operation
      handler.reset()
      handler.find_all_by_prop_val(root, 'pos', 'NNP')
      handler.find_all_by_prop_val(root, 'pos', 'NN')
      entity_nodes = handler.found_nodes

      for entity_node in entity_nodes:
        #Find single-token entities
        if ((len(str(entity_node)) > 1) and (self.lm.lemmatize(str(entity_node), 'v') not in self.known_operations and
          str(entity_node) not in self.known_conditions and str(entity_node) not in self.known_entity_exceptions and
          not isinstance(str(entity_node), float) or str(entity_node) in self.database_chemicals)):
          extracted_data['entities'].append(unicode(entity_node))

        #Find multi-token entities
        #TODO: Fix matching brackets, spacing issues
        if entity_node.data['pos'] == 'NNP' or entity_node.data['pos'] == 'CD':
          start_token = entity_node
          possible_entity = []
          entity_stop_scan = False
          id_ahead = 0
          allowed_np_pos = ['NNP', 'NN', '-LRB-', '-RRB-', 'CD', 'HYPH']

          while(not entity_stop_scan):
            handler.find_node_by_id(true_root, int(start_token.data['id']) + int(id_ahead))
            next_token = handler.found_node

            if next_token is not None:
              if next_token.data['pos'] in allowed_np_pos:
                possible_entity.append(str(next_token))
                handler.reset()
                id_ahead += 1
              else:
                if len(possible_entity) > 1:
                  multitoken_entity = ''.join(possible_entity)
                  if multitoken_entity not in extracted_data['entities']:
                    extracted_data['entities'].append(unicode(multitoken_entity))
                entity_stop_scan = True
            else:
              entity_stop_scan = True



      #Search for conditions associated with this operation
      handler.reset()
      for condition in self.known_conditions:
        handler.find_all_by_name(root, condition)

      condition_nodes = handler.found_nodes

      #Get numeric conditions (e.g. 500C or 300rpm)
      handler.reset()
      handler.find_all_children(root)

      numeric_condition_nodes = []
      for node in handler.found_nodes:
        handler.find_node_by_id(true_root, int(node.data['id']) + 1)
        if (str(handler.found_node) in self.known_numeric_conditions):
          numeric_condition_nodes.append(unicode(node) + unicode(handler.found_node))

      for condition_node in condition_nodes:
        if unicode(condition_node) not in extracted_data['conditions']:
          extracted_data['conditions'].append(unicode(condition_node))

      extracted_data['conditions'].extend(numeric_condition_nodes)

      extracted_data_list.append(extracted_data)

    for child_node in root.children:
      self.extract_from_tree(true_root, child_node, extracted_data_list)

  def extract_target_material_from_parsed_abstract(self, root, extracted_targets):
    lem_root_v = self.lm.lemmatize(str(root).lower(), 'v')
    lem_root_n = self.lm.lemmatize(str(root).lower(), 'n')
    handler = NodeHandler()

    #Root verb approach doesn't work with 'synthesis' or whatever is a noun...
    #TODO: FIX THIS!!!
    if (lem_root_v in self.known_target_material_verbs or
    lem_root_n in self.known_target_material_nouns):

      #Search for target materials associated with this operation
      handler.reset()
      handler.find_all_by_prop_val(root, 'phrase', 'dobj')
      handler.find_all_by_prop_val(root, 'phrase', 'nsubjpass')
      handler.find_all_by_prop_val(root, 'phrase', 'conj')
      handler.find_all_by_prop_val(root, 'phrase', 'nsubj')
      target_nodes = handler.found_nodes

      for target_node in target_nodes:
        if (match('^([(]?([A-Z][a-z]?([0-9]*[+-]?)?)+[)]?[0-9]*[:/]?)+$', str(target_node)) is not None
        and str(target_node) not in extracted_targets):
          extracted_targets.append(unicode(target_node))

  def extract_target_material_from_abstract(self, abstract):
    abstract_sentences = sent_tokenize(abstract)
    found_words = []

    for sentence in abstract_sentences:
      abstract_words = pos_tag(word_tokenize(sentence))

      for i, word_tuple in enumerate(abstract_words):
        if (self.lm.lemmatize(word_tuple[0], 'v') in self.known_target_material_verbs or
        self.lm.lemmatize(word_tuple[0], 'n') in self.known_target_material_nouns):
          #Start doing a proximity search for the target material
          candidate_words = []

          for offset in range(1, len(abstract_words)):
            before_word = None
            after_word = None

            if i - offset >= 0:
              before_word = abstract_words[i - offset]
            if i + offset < len(abstract_words):
              after_word = abstract_words[i + offset]

            allowed_tags = ['NN', 'NNP', 'NNS', 'NP', 'NN-t1', 'NP-t1', 'N', 'X']
            if before_word is not None:
              if (before_word[1] in allowed_tags) and (len(before_word[0]) > 3):
                candidate_words.append(before_word[0])
            if after_word is not None:
              if (after_word[1] in allowed_tags) and (len(after_word[0]) > 3):
                candidate_words.append(after_word[0])

          for word in candidate_words:
            if (match('^([(]?([A-Z][a-z]?([0-9]*[+-]?)?)+[)]?[0-9]*[:/]?)+$', word) is not None
            and word not in found_words):
              found_words.append(unicode(word))

    if len(found_words) == 0:
      return [unicode('MATERIALS_NOT_FOUND')]
    else:
      return found_words


class SynthRouteExtractorTester(SynthRouteExtractor):

  def load_sentence_file(self, filename):
    json_sentences = {}
    self.gold_data = [] #Gold standard test data
    self.gold_sentences = []

    with open(filename) as f:
      json_sentences = loads(f.read())

    for sentence in json_sentences['sentences']:
       self.gold_sentences.append(sentence['sentence'])
       self.gold_data.append(sentence['extracted_data'])

    print 'Test sentence data loaded'

  def load_abstract_file(self, filename):
    json_abstracts = {}
    self.gold_abstracts = []
    self.gold_target_materials = []

    with open(filename) as f:
      json_abstracts = loads(f.read())

    for abstract_data in json_abstracts:
      self.gold_abstracts.append(abstract_data['abstract'])
      self.gold_target_materials.append(abstract_data['target_materials'])

    print 'Test abstract data loaded'

  def parse_all_sentences(self):

    self.parse_trees = []
    for sentence in self.gold_sentences:
      roots = self.parser.parse_sentence(sentence)
      if not roots == 1: self.parse_trees.append(roots)

  def extract_sentence_data(self):
    self.extracted_data = []
    for i, tree_list in enumerate(self.parse_trees):
      self.extracted_data.append([])
      for tree in tree_list:
        self.extracted_data[i].append([])
        self.extract_from_tree(tree, tree, self.extracted_data[i][-1])

    for i, inp_para in enumerate(self.extracted_data):
      for j, ext_sentence in enumerate(inp_para):
        self.extracted_data[i][j] = sorted(ext_sentence, key=lambda op_data: int(op_data['sent_id']))

  def extract_target_materials(self):
    self.extracted_targets = []
    for abstract in self.gold_abstracts:
      self.extracted_targets.append(self.extract_target_material_from_abstract(abstract))
    # for abstract in self.gold_abstracts:
    #   extracted_targets = []
    #   roots = self.parser.parse_sentence(abstract)
    #   for root in roots:
    #     self.extract_target_material_from_parsed_abstract(root, extracted_targets)
    #   self.extracted_targets.append(extracted_targets)

  def compute_extracted_target_materials_accuracy(self):
    targets_pos, targets_guesses, targets_tot = 0.0, 0.0, 0.0

    for extracted_target_list, target_materials in zip(self.extracted_targets, self.gold_target_materials):
      targets_tot += len(target_materials)
      targets_guesses += len(extracted_target_list)
      for extracted_target in extracted_target_list:
        if extracted_target in target_materials:
          targets_pos += 1

    #precision, recall
    return (round(targets_pos / targets_guesses, 2), round(targets_pos / targets_tot, 2))

  def compute_extracted_data_accuracy(self):
    operations_pos, operations_guesses, operations_tot = 0.0, 0.0, 0.0
    entity_relations_pos, entity_relations_guesses, entity_relations_tot = 0.0, 0.0, 0.0
    entity_pos, entity_guesses, entity_tot = 0.0, 0.0, 0.0
    condition_relations_pos, condition_relations_guesses, condition_relations_tot = 0.0, 0.0, 0.0
    condition_pos, condition_guesses, condition_tot = 0.0, 0.0, 0.0
    gold_operations = []
    gold_entities = []
    gold_flat_entities = []
    gold_conditions = []
    gold_flat_conditions = []

    #Compare test predictions vs gold standards
    for gold_sentence in self.gold_data:
        gold_operations.append([str(s['operation']) for s in gold_sentence])
        operations_tot += len(gold_operations[-1])
        gold_entities.append([])
        gold_conditions.append([])
        gold_flat_entities.append([])
        gold_flat_conditions.append([])
        for extracted_data in gold_sentence:
          gold_flat_entities[-1].extend(extracted_data['entities'])
          gold_flat_conditions[-1].extend(extracted_data['conditions'])
          gold_entities[-1].append(extracted_data['entities'])
          gold_conditions[-1].append(extracted_data['conditions'])
          entity_tot += len(extracted_data['entities'])
          condition_tot += len(extracted_data['conditions'])
          entity_relations_tot += len(extracted_data['entities'])
          condition_relations_tot += len(extracted_data['conditions'])

    for i, inp_sentence in enumerate(self.extracted_data):
      for ext_sentence in inp_sentence:
        for j, operation_data in enumerate(ext_sentence):
          operations_guesses += 1
          if str(operation_data['operation']) in gold_operations[i]:
            operations_pos += 1
          for entity in operation_data['entities']:
            entity_relations_guesses += 1
            entity_guesses += 1
            if str(entity) in gold_flat_entities[i]:
              entity_pos += 1
            if j < len(gold_conditions[i]):
              if str(entity) in gold_entities[i][j]:
                entity_relations_pos += 1
          for condition in operation_data['conditions']:
            condition_relations_guesses += 1
            condition_guesses += 1
            if str(condition) in gold_flat_conditions[i]:
              condition_pos += 1
            if j < len(gold_conditions[i]):
              if str(condition) in gold_conditions[i][j]:
                condition_relations_pos += 1


    #op precision, op recall, ent precision, ent recall, cond precision, cond recall
    return (round(operations_pos / operations_guesses, 2), round(operations_pos / operations_tot, 2),
    round(entity_relations_pos / entity_relations_guesses, 2), round(entity_relations_pos / entity_relations_tot, 2),
    round(entity_pos / entity_guesses, 2), round(entity_pos / entity_tot, 2),
    round(condition_relations_pos / condition_relations_guesses, 2), round(condition_relations_pos / condition_relations_tot, 2),
    round(condition_pos / condition_guesses, 2), round(condition_pos / condition_tot, 2))
